part of viltage;

class Collidable {
  int x, y;
  Entity entity;
  
  /** Creates a new Collidable object.
   *  Pass an Entity object if you want this Collidable to move along with the Entity.
   * 
   *      new Collidable(x, y, entity);
   * 
   *  Passing an Entity also means that x and y values are not absolute positions, but offsets from Entity's position.
   */
  Collidable(this.x, this.y, [this.entity = null]) {
    if(entity != null) entity._collidables.add(this);
  }
  
  /** Use this method only if you want to unbind a Collidable object from an Entity */  
  remove() {
    if(entity != null) entity._collidables.remove(this);
  }
}

class Collision {
  Collidable collidable1;
  Collidable collidable2;
  /** Directions are in respect to collidable1 object. **/
  bool left, right, up, down, overlap; 
  
  Collision(this.collidable1, this.collidable2, this.left, this.right, this.up, this.down, this.overlap);
  
  /** This method accepts any combination of Collidables or Entities as an arguments.
   *  Returns a list of Collision objects (list is empty if no collision occured). 
   * 
   *     Collision.check(entity, collidable);
   *     Collision.check(entity, entity);
   *     Collision.check(collidable, collidable);
   */
  static List<Collision> check(Object obj1, Object obj2) {
    List<Collision> collisions = new List<Collision>();
    
    if(obj1 is Collidable && obj2 is Collidable) {
      Collision c = checkCollidables(obj1, obj2);
      if(c!=null) collisions.add(c);
    } else if(obj1 is Entity && obj2 is Entity) {
      for(int i = 0; i < obj1._collidables.length; i++) {
        for(int j = 0; j < obj2._collidables.length; j++) {
          Collision c = checkCollidables(obj1._collidables[i], obj2._collidables[j]);
          if(c!=null) collisions.add(c);
        }
      }
    } else if((obj1 is Entity && obj2 is Collidable)) {
      for(int i = 0; i < obj1._collidables.length; i++) {
        Collision c = checkCollidables(obj1._collidables[i], obj2);
        if(c!=null) collisions.add(c);
      }
    } else if((obj1 is Collidable && obj2 is Entity)) {
      for(int i = 0; i < obj2._collidables.length; i++) {
        Collision c = checkCollidables(obj1, obj2._collidables[i]);
        if(c!=null) collisions.add(c);
      }
    }
    
    return collisions;
  }
  
  /** This method accepts only two Collidables and returns a Collision object (null if no collision).
   *  Use this if you want to avoid lists from more universal method (Collision.check(..)).
   * 
   *     Collision.checkCollidables(collidable, collidable);
   */
  static Collision checkCollidables(Collidable c1, Collidable c2) {
    bool left =    (c1.x+(c1.entity==null?0:c1.entity.x))+1 == (c2.x+(c2.entity==null?0:c2.entity.x)) && (c1.y+(c1.entity==null?0:c1.entity.y)) == (c2.y+(c2.entity==null?0:c2.entity.y));
    bool right =   (c1.x+(c1.entity==null?0:c1.entity.x))-1 == (c2.x+(c2.entity==null?0:c2.entity.x)) && (c1.y+(c1.entity==null?0:c1.entity.y)) == (c2.y+(c2.entity==null?0:c2.entity.y));
    bool up =      (c1.y+(c1.entity==null?0:c1.entity.y))+1 == (c2.y+(c2.entity==null?0:c2.entity.y)) && (c1.x+(c1.entity==null?0:c1.entity.x)) == (c2.x+(c2.entity==null?0:c2.entity.x));
    bool down =    (c1.y+(c1.entity==null?0:c1.entity.y))-1 == (c2.y+(c2.entity==null?0:c2.entity.y)) && (c1.x+(c1.entity==null?0:c1.entity.x)) == (c2.x+(c2.entity==null?0:c2.entity.x));
    bool overlap = (c1.x+(c1.entity==null?0:c1.entity.x))   == (c2.x+(c2.entity==null?0:c2.entity.x)) && (c1.y+(c1.entity==null?0:c1.entity.y)) == (c2.y+(c2.entity==null?0:c2.entity.y));
    if(left || right || up || down || overlap) return new Collision(c1, c2, left, right, up, down, overlap);
    else return null;
  }
}